/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider;

import java.util.ArrayList;

import javax.xml.namespace.QName;
import org.w3c.dom.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;
import org.globus.wsrf.impl.timer.TimerManagerImpl;

import org.globus.mds.usefulrp.types.ResourcePropertyProviderConfig;

/** 
    Abstract base class for ResourcePropertyProvider implementations.
 
    Initializes a <code>ReflectionResourceProperty</code> with the QName 
    pecified in the <code>ResourcePropertyProviderConfig</code> constructor.  
    Also creates a <code>TimerManagerImpl</code> object which derived classes 
    may use to schedule the <code>ResourcePropertyProviderTasks</code> which 
    will periodically populate the values of the 
    <code>ReflectionResourceProperty</code>.
*/

public abstract class BaseResourcePropertyProvider extends ReflectionResourceProperty 
{
    private static Log logger =
            LogFactory.getLog(BaseResourcePropertyProvider.class.getName());
        
    protected ResourcePropertyProviderConfig config;
    protected TimerManagerImpl timerManager;
    
   /**  
    * @param config the <code>ResourcePropertyProviderConfig</code> the 
    *              configuration parameters for this instance of the 
    *              <code>GLUEResourceProperty</code> provider.
    *
    * @throws java.lang.Exception passes along any exception encountered
    *    during construction of underlying ReflectionResourceProperty.
    */
    public BaseResourcePropertyProvider(ResourcePropertyProviderConfig config) throws Exception
    {
        super(new SimpleResourcePropertyMetaData(config.getResourcePropertyName()));
        this.config = config;
        setObject(this);
        setPropertyName("DynamicValue");
        
        this.timerManager = new TimerManagerImpl();
         
        initialize();
    }
    
    public ResourcePropertyProviderConfig getConfig() {
        return this.config;
    }
    
    public TimerManagerImpl getTimerManager() {
        return this.timerManager;
    }
    
    protected void finalize() {
        if (this.timerManager != null) {
            this.timerManager.stop();
        }
    }
    
}