/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider;

import java.util.ArrayList;

import javax.xml.namespace.QName;
import org.w3c.dom.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.impl.timer.TimerManagerImpl;

import org.globus.mds.usefulrp.types.ResourcePropertyProviderConfig;

import org.globus.mds.glue.GLUECERPType;
import org.globus.mds.glue.ClusterType;
import org.globus.mds.glue.ComputingElementType;
import org.globus.mds.glue.batchprovider.ClusterCollectionType;

/** 
 *  An implementation of the GLUE Resource Property using the Resource Property 
 *  Provider API. <p>
 *
 *  This class represents GLUE 1.1 XML data as a resource property.  
 *  The structure consists of an RP based on the <code>GLUECERPType</code> 
 *  generated from XSD schema.<p>
 *
 *  The implementation uses <code>ResourcePropertyElementProducers</code> and 
 *  <code>ResourcePropertyElementTransforms</code> to generate and post-process 
 *  data obtained from underlying cluster and schedluer information providers, 
 *  which can include external systems such as Ganglia, Clumon, PBS, LSF, etc. 
 */

public class GLUEResourceProperty extends BaseResourcePropertyProvider 
{
    private static Log logger =
            LogFactory.getLog(GLUEResourceProperty.class.getName());
            
    private TransformElementListener[] listeners;
    
    private GLUECERPType glue = new GLUECERPType();
    private ClusterType[] clusterType = new ClusterType[]{};
    private ComputingElementType[] ceType = new ComputingElementType[]{};
    
   /**  
    * @param config the <code>ResourcePropertyProviderConfig</code> the 
    *              configuration parameters for this instance of the 
    *              <code>GLUEResourceProperty</code> provider.
    *
    * @throws java.lang.Exception passes along any exception encountered
    *    during construction of underlying ReflectionResourceProperty.
    */
    public GLUEResourceProperty(ResourcePropertyProviderConfig config) 
        throws Exception
    {
        super(config);        
    }
    
    public void initialize() throws Exception 
    {
        super.initialize();
        
        this.getTimerManager().setPoolSize(2);
        
        this.listeners = 
            TransformElementListener.scheduleListeners(this.getConfig(), 
                this.getTimerManager());
    }
    
    private void setGLUERPValue (Object obj) 
    {
        if (obj == null) {
            if (logger.isDebugEnabled()) {
                logger.debug("Null value returned from an element producer");
            }
            return;
        }
        
        if (obj instanceof ClusterType[])
        {
            this.glue.setCluster((ClusterType[])obj);
        }
        else if (obj instanceof ComputingElementType[]) 
        {
            this.glue.setComputingElement((ComputingElementType[])obj);
        }
        else if (obj != null)
        {
            logger.warn("Invalid object type returned from an element producer: " + 
                obj.getClass().getName());
        }
    }
    
    public GLUECERPType getDynamicValue() 
    {        
        this.glue.setCluster(this.clusterType);
        this.glue.setComputingElement(this.ceType);
        
        if (this.listeners == null)  {
            return this.glue;
        }
            
        if (this.listeners.length > 0) 
        {
            Object obj = listeners[0].getObject();
            this.setGLUERPValue(obj);
        }

        if (this.listeners.length > 1) 
        {
            Object obj = listeners[1].getObject();
            this.setGLUERPValue(obj);
        }
        
        if (this.listeners.length > 2) {
            Object obj = listeners[2].getObject();            
            logger.warn("Unexpected additional object type returned from an element producer (this and all subsequent elements will not be published): " + 
                obj.getClass().getName());
        }
        
        return this.glue;
    }         

}