/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider;

import java.util.ArrayList;

import javax.xml.namespace.QName;
import org.w3c.dom.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.utils.XMLUtils;

import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.w3c.dom.Document;

import java.io.FileWriter;
import java.io.FileInputStream;

import org.globus.mds.usefulrp.types.ResourcePropertyProviderConfig;
import org.globus.mds.usefulrp.types.ResourcePropertyProviderConfigArray;
import org.globus.mds.usefulrp.types.ResourcePropertyElementProducerConfig;

/** 
 * Utility class for writing config files for GLUECERP
 */

public class GLUEResourcePropertyConfig 
{
    private static Log logger =
            LogFactory.getLog(GLUEResourcePropertyConfig.class.getName());
            
    public GLUEResourcePropertyConfig() {
   
    }
    
    public static void main(String args[]) throws Exception
    {              
        String fileName = "gluerp-config.xml";
        String schedulerInfoProviderName = null; 
        String clusterInfoProviderName = null;
        
        if ((args == null) || (args.length < 2)) {
            System.out.println();
            System.out.println("ERROR: Not enough arguments - must include at least one valid scheduler and one valid cluster information provider name or \"none\"");
            usage();
            return;
        }                
        
        if (args.length > 0) {        
            schedulerInfoProviderName = args[0];
        }        
        if (args.length > 1) {        
            clusterInfoProviderName = args[1];
        }
        if (args.length > 2) {
            fileName = args[2]; 
        }       
               
        try {
            GLUEResourcePropertyConfig.generateDefaultConfig(
                fileName, schedulerInfoProviderName, clusterInfoProviderName
            );
        }
        catch (Exception e) {
            System.out.println();            
            System.out.println(e);
            usage();
            return;
        }
        
        System.out.println();
        System.out.println("Successfuly wrote configuration output file to: " + fileName);
        System.out.println();
    }

    protected static void usage() {
        System.out.println();
        System.out.println("Usage: [scheduler info provider name | keyword none] [cluster info provider name | keyword none] [outputFile - defaults to gluerp-config.xml in the current dir]" );
        System.out.println();
        System.out.println("Example: pbs ganglia" );
        System.out.println("Example: none clumon" );
        System.out.println("Example: fork none $GLOBUS_LOCATION/etc/gram-service-Fork/gluerp-config.xml" );
        System.out.println();
    }
    
    public static void generateDefaultConfig(String fileName,
                                             String schedulerInfoProviderName,
                                             String clusterInfoProviderName) 
        throws Exception
    {
        if (fileName == null) {
            throw new IllegalArgumentException("Output filename must not be null!");
        }
        
        if ((schedulerInfoProviderName == null) && (clusterInfoProviderName == null)) {
            throw new IllegalArgumentException("Must provide at least one schedulerInfoProviderName or clusterInfoProviderName (or both)");
        } 
        
        ResourcePropertyProviderConfigArray config = 
            new ResourcePropertyProviderConfigArray();
        
        ResourcePropertyProviderConfig[] configArray = 
            new ResourcePropertyProviderConfig[] { 
                new ResourcePropertyProviderConfig() };
        
        configArray[0].setResourcePropertyImpl(
            "org.globus.mds.usefulrp.rpprovider.GLUEResourceProperty"
        );

        configArray[0].setResourcePropertyName(
            new QName("http://mds.globus.org/glue/ce/1.1","mds:GLUECE"));

        ArrayList producers = new ArrayList();        

        // process any scheduler provider configuration
        if (schedulerInfoProviderName == null) {
            if (logger.isDebugEnabled()) {
                logger.debug("No cluster information provider name specified");
            }
        } 
        else if (schedulerInfoProviderName.equalsIgnoreCase("none"))
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Scheduler information provider set to none");
            }            
        }
        else
        {
            ResourcePropertyElementProducerConfig schedulerProducer = 
                new ResourcePropertyElementProducerConfig();    
            
            schedulerProducer.setClassName(
                "org.globus.mds.usefulrp.rpprovider.producers.SchedulerInfoElementProducer"
            );

            String baseScriptArg = "libexec/globus-scheduler-provider-";
            schedulerProducer.setArguments(
                new String[] {
                    baseScriptArg + schedulerInfoProviderName.toLowerCase()
                }
            );          

            schedulerProducer.setPeriod(new Integer(300));

            schedulerProducer.setTransformClass(
                "org.globus.mds.usefulrp.rpprovider.transforms.GLUESchedulerElementTransform");

            schedulerProducer.setTransformArguments(new String[] {} );

            producers.add(schedulerProducer);
        }

        // process any cluster provider configuration 
        if (clusterInfoProviderName == null) {
            if (logger.isDebugEnabled()) {
                logger.debug("No cluster information provider name specified");
            }
        } 
        else if (clusterInfoProviderName.equalsIgnoreCase("none"))
        {
            if (logger.isDebugEnabled()) {
                logger.debug("Cluster information provider set to none");
            }
        }
        else if (clusterInfoProviderName.equalsIgnoreCase("ganglia"))  
        {
	    ResourcePropertyElementProducerConfig clusterProducer = 
	   	new ResourcePropertyElementProducerConfig();
            
	    clusterProducer.setClassName(
                "org.globus.mds.usefulrp.glue.GangliaElementProducer"
            );

            clusterProducer.setArguments(
                new String[] { "localhost", "8649" }
            ); 

            clusterProducer.setPeriod(new Integer(300));

            clusterProducer.setTransformClass(
                "org.globus.mds.usefulrp.rpprovider.transforms.GLUEComputeElementTransform");

            clusterProducer.setTransformArguments(new String[] {} );
            
            producers.add(clusterProducer);
        } 
        else if (clusterInfoProviderName.equalsIgnoreCase("clumon")) 
        {
	    ResourcePropertyElementProducerConfig clusterProducer = 
	   	new ResourcePropertyElementProducerConfig();
            
            clusterProducer.setClassName(
                "org.globus.mds.usefulrp.rpprovider.producers.URLElementProducer"
            );

            clusterProducer.setArguments(
                new String[] {
                    "http://localhost/glue.php"
                }
            );

            clusterProducer.setPeriod(new Integer(300));

            clusterProducer.setTransformClass(
                "org.globus.mds.usefulrp.rpprovider.transforms.GLUEComputeElementTransform");

            clusterProducer.setTransformArguments(
                new String[] {"./etc/globus_wsrf_mds_usefulrp/cluster_config.xml"} );
            
            producers.add(clusterProducer);            
        }
        else if (clusterInfoProviderName.equalsIgnoreCase("nagios"))
        {
	    ResourcePropertyElementProducerConfig clusterProducer = 
	   	new ResourcePropertyElementProducerConfig();
            
	    clusterProducer.setClassName(
                "org.globus.mds.usefulrp.rpprovider.producers.ExternalProcessElementProducer"
            );

            String globusLoc = System.getProperty("GLOBUS_LOCATION");                    
            clusterProducer.setArguments(
                new String[] { ((globusLoc == null) ? "." : globusLoc) + 
                                "/libexec/globus-mds-cluster-nagios" } ); 

            clusterProducer.setPeriod(new Integer(300));
            
            producers.add(clusterProducer);            
        }        
        else 
        {
            throw new IllegalArgumentException(
                "Unknown cluster information provider type: " + 
                    clusterInfoProviderName);
        }

        configArray[0].setResourcePropertyElementProducers(
            (ResourcePropertyElementProducerConfig[])
                producers.toArray(new ResourcePropertyElementProducerConfig[]{})
        );                      
        
        config.setConfigArray(configArray);
        
        String header = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n";
        
        String comment = 
            "<!-- Configuration file for the MDS GLUECE resource property provider -->\n\n";
            
        Element element = 
            ObjectSerializer.toElement(config,
                config.getTypeDesc().getXmlType());
        
        String output = 
            header + comment + 
                XMLUtils.PrettyDocumentToString(element.getOwnerDocument());
        
	if (logger.isDebugEnabled()) {
            logger.debug(output);    
        }
                
        FileWriter filewriter = null;
        try {
            filewriter = new FileWriter(fileName);
            filewriter.write(output);
        } catch (Exception e) {
            throw e;
        } finally {
            try {
                if (filewriter != null) {
                    filewriter.close();
                }
            } catch (Exception e) {}
        }
    }      

    
    
}
