/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider.producers;

import org.globus.wsrf.utils.XmlUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.mds.usefulrp.rpprovider.ResourcePropertyElementProducer;

/**
 *  A <code>ResourcePropertyElementProducer</code> which executes a external 
 *  process, then attempts to parse the STDOUT stream of that process into a 
 *  valid XML document.<p>
 *
 *  The argument vector passed to the getElement method of this class is passed
 *  unmodified to the child process as its argument vector.<p>
 *  
 *  <b>Note:</b> This implementation will not work with .CMD or .BAT script 
 *               files on MS-Windows platforms, though it will function with 
 *              normal executable (.EXE) files on such platforms.<p>
*/

public class ExternalProcessElementProducer 
    implements ResourcePropertyElementProducer
{

    private static Log logger =
            LogFactory.getLog(ExternalProcessElementProducer.class.getName());
    
    /** Creates a new instance of ExternalProcessElementProducer */
    public ExternalProcessElementProducer() {
        
    }
    
    public Element getElement(String[] args) throws Exception
    {
        if ((args == null) || ((args != null) && (args.length == 0))) {
             throw new IllegalArgumentException("Arguments cannot be null!");
        }
        
        Document doc = null;
        Process process = null;
        Runtime runtime = Runtime.getRuntime();  
        
        try 
        {
            process = runtime.exec(args);           
        } 
        catch (Exception e) 
        {
            throw new Exception("Process execution error: " + e);
        }
        
        try 
        {                   
            if (process != null) {
                doc = XmlUtils.newDocument(process.getInputStream());
                if (logger.isDebugEnabled()) {
                    logger.debug("Document parsing successful:\n" + 
                                 XmlUtils.toString(doc));
                }
            }         
        }
        catch (Exception e)
        {
            throw new Exception("Exception while parsing child process stdout into valid XML document: " + e);
        }
        finally
        {
            int exit = 0;
            try {
                exit = process.exitValue();
            }
            catch (IllegalThreadStateException itse) 
            {
                if (logger.isDebugEnabled()) {
		    logger.debug("Waiting for child process to terminate: " + args[0]);
		}
                exit = process.waitFor(); 
                if (logger.isDebugEnabled()) {
		    logger.debug("Child process " + args[0] + 
                            " exited with return code: " + 
                            Integer.toString(exit));
		}
            }
        }
        return doc.getDocumentElement();                
    }
      
}
