/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider.producers;

import org.globus.wsrf.utils.XmlUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.InputStream;
import java.io.FileInputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.mds.usefulrp.rpprovider.ResourcePropertyElementProducer;

/** 
 * A ResourcePropertyElementProducer which generates an DOM Element by fetching
 * an XML file from the specified URL parameter.  This URL is generally expected 
 * to be a URL to an HTTP(S) server, and the result of the HTTP GET request on 
 * this URL is generally expected to be a well-formed XML document.
*/

public class URLElementProducer implements ResourcePropertyElementProducer
{

    private static Log logger =
            LogFactory.getLog(URLElementProducer.class.getName());
    
    /** Creates a new instance of URLElementProducer */
    public URLElementProducer() {
        
    }
    
    public Element getElement(String[] args) throws Exception
    {
        if ((args == null) || ((args != null) && (args.length == 0))) {
             throw new IllegalArgumentException("Arguments cannot be null!");
        }
        
        Document doc = null;
        String url = args[0];
        String username = null;
        String password = null;
        
        if (args.length > 1) {
            username = args[1];
        }
        
        if (args.length > 2) {
            password = args[2];
        }
        
        try
        {
            doc = XmlUtils.newDocument(url, username, password);
        }
        catch (Exception e)
        {
            throw new Exception("Could not parse output from " + url);
        }
        return doc.getDocumentElement();                    
    }
      
}
