/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider.transforms;

import org.globus.wsrf.utils.XmlUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.globus.mds.glue.ClusterType;
import org.globus.mds.glue.ComputingElementType;
import org.globus.mds.glue.InfoType;
import org.globus.mds.glue.JobType;
import org.globus.mds.glue.StateType;
import org.globus.mds.glue.PolicyType;
import org.globus.mds.glue.batchprovider.ClusterCollectionType;

import org.globus.mds.glue.config.ClusterConfiguration;

import org.globus.mds.usefulrp.rpprovider.ResourcePropertyElementTransform;

/** 
*/

public class GLUEComputeElementTransform 
    implements ResourcePropertyElementTransform
{
    private static Log logger =
            LogFactory.getLog(GLUEComputeElementTransform.class.getName());

    ClusterConfiguration config = null;
    long configLastModified = 0;
    
    /** Creates a new instance of GLUEComputeElementTransform */
    public GLUEComputeElementTransform() {
        
    }
    
    private void rescanConfig(String configFile) throws Exception
    {
        File file  = new File(configFile);
        if (this.configLastModified == 0) {
            this.configLastModified = file.lastModified();
        } else if (this.configLastModified >= file.lastModified()) {
            return;            
        } 
        this.configLastModified = file.lastModified();        

        if (logger.isDebugEnabled()) {
            logger.debug("Processing subcluster definition configuration file");
        }          
        
        this.config = GLUEXMLOrganizer.parseConfig(configFile);
        
        if (logger.isDebugEnabled()) {
            logger.debug("Processed a valid GLUE subcluster definition configuration file");
        }        
    }
    
    public Object transformElement(Element source, String[] args) 
        throws Exception
    {
        if (logger.isDebugEnabled()) {
            logger.debug("Input document XML: " + 
                ObjectSerializer.toString(source));
        }  
                    
        Element resultElement = source;

        if ((args != null) && (args.length > 0))
        {
            try 
            {
                // NOTE:
                // args[0] is the config file.
                // args[1] is an optional string which will be
                // appended to the hostnames to make them unique
                this.rescanConfig(args[0]);
                
                if (this.config != null)
                {
                    Document sourceDoc = source.getOwnerDocument();

                    Document destDoc = 
                        GLUEXMLOrganizer.reorganizeGLUEDocument(
                            this.config, sourceDoc,
                            ((args.length > 1) ? args[1] : null));

                    resultElement = destDoc.getDocumentElement();
                    
                    if (logger.isDebugEnabled()) {
                        logger.debug("Successfully reorganized GLUE document XML: " + 
                            ObjectSerializer.toString(resultElement));
                    }                    
                }
            }
            catch (Exception e)
            {
                logger.warn("Exception while attempting to collate GLUE information: " + e);
            }
        }
             
        // for now we are going to deserialize this into a ClusterCollectionType
        // automatically, just like the GT4.0.x GlueResourceProperty.clusterListener
        ClusterCollectionType clusters = null;
        try {
            clusters =
                (ClusterCollectionType) ObjectDeserializer.toObject(
                    resultElement, ClusterCollectionType.class);
        }
        catch (Exception e)
        {
            logger.warn("Exception while trying to deserialize GLUE input: " + e);
        }
        
        if (logger.isDebugEnabled()) {
            logger.debug(
                "Successfully deserialized transform input into ClusterCollectionType object: " + 
                    ObjectSerializer.toString(clusters));
        }     
        
        
        if (clusters != null) {
            return clusters.getCluster();
        } else {
            return null;
        }            
    }
}
