/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider.transforms;

import org.globus.wsrf.utils.XmlUtils;

import org.w3c.dom.Element;
import org.w3c.dom.Document;

import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.apache.axis.message.MessageElement;

import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.globus.mds.glue.ClusterType;
import org.globus.mds.glue.ComputingElementType;
import org.globus.mds.glue.InfoType;
import org.globus.mds.glue.JobType;
import org.globus.mds.glue.StateType;
import org.globus.mds.glue.PolicyType;
import org.globus.mds.glue.batchprovider.SchedulerType;
import org.globus.mds.glue.batchprovider.QueueType;

import org.globus.mds.glue.config.HostExtType;

import org.globus.mds.usefulrp.rpprovider.ResourcePropertyElementTransform;

/** 
*/

public class GLUESchedulerElementTransform 
    implements ResourcePropertyElementTransform
{

    private static Log logger =
            LogFactory.getLog(GLUESchedulerElementTransform.class.getName());
    
    /** Creates a new instance of GLUEComputeElementTransform */
    public GLUESchedulerElementTransform() {

    }

    public Object transformElement(Element source, String[] args)
        throws Exception
    {
        ComputingElementType[] ces = null;
        InfoType additionalInfo = null;

        try 
        {
            if ((args != null) && (args.length > 0))
            {
                try
                {
                    // NOTE:
                    // args[0] is an optional scheduler file that can
                    // contain InfoType values that will override any
                    // incoming from the scheduler provider.
                    // args[1] is an optional string which will be
                    // appended to the hostnames (of the
                    // TeraGridHostExtensions, if specified) to make
                    // them unique.
                    // if args[1] is desired and args[0] is not, just
                    // fill in a dummy variable for args[0] and mind
                    // the warning
                    DocumentBuilderFactory factory =
                        DocumentBuilderFactory.newInstance();
                    factory.setIgnoringComments(true);
                    factory.setCoalescing(true);
                    factory.setNamespaceAware(true);
                    factory.setValidating(false);

                    DocumentBuilder db = factory.newDocumentBuilder();
                    Document doc = db.parse(new File(args[0]));
                    Element elem = doc.getDocumentElement();

                    ComputingElementType computingElem =
                        (ComputingElementType)ObjectDeserializer.toObject(
                            elem, ComputingElementType.class);

                    if (computingElem != null)
                    {
                        additionalInfo = computingElem.getInfo();
                    }
                }
                catch (Exception e)
                {
                    logger.warn("Exception while reading specified " +
                                "GLUEScheduler file: " + e);
                }
            }

            SchedulerType schedulerOutput = (SchedulerType)
                ObjectDeserializer.toObject(source, SchedulerType.class);

            if (schedulerOutput.getQueue() == null) {
                throw new Exception("Batch provider generated no useful information.");
            }

            if (logger.isDebugEnabled()) {
                logger.debug("Converting batch system information to " +
                             "GLUE ComputeElement schema.");
            }

            if (logger.isDebugEnabled()) {
                logger.debug("There are "+
                             schedulerOutput.getQueue().length +
                             " queues.");
            }

            if (logger.isDebugEnabled() &&
                (schedulerOutput.get_any() != null)) {
                logger.debug(
                    "There appears to be some Any Extensions present.");
            }

            ces = new ComputingElementType[schedulerOutput.getQueue().length];

            for(int i = 0; i < schedulerOutput.getQueue().length; i++)
            {
                ComputingElementType ce = new ComputingElementType();
                InfoType info = schedulerOutput.getInfo();
                JobType[] jobs = schedulerOutput.getJob();
                QueueType queue = schedulerOutput.getQueue()[i];
                HostExtType[] tgHostExts = null;
                MessageElement[] msgElems = schedulerOutput.get_any();

                // only add teragrid host extensions (if present) to
                // the first computing element since they can contain
                // large amounts of data that shouldn't be duplicated
                if ((msgElems != null) && (i == 0))
                {
                    try {
                        Element tgElem = null;
                        int numMsgElems = msgElems.length;
                        HostExtType[] newHostExts =
                            new HostExtType[numMsgElems];
                        MessageElement[] newMsgElems =
                            new MessageElement[numMsgElems];

                        for(int j = 0; j < numMsgElems; j++)
                        {
                            HostExtType cur = null;
                            tgElem = msgElems[j].getAsDOM();

                            if ((tgElem != null) &&
                                (GLUEXMLOrganizer.removeNS(
                                    tgElem.getTagName()).equals(
                                        "TeraGridHostExtensions")))
                            {
                                cur = (HostExtType)
                                    ObjectDeserializer.toObject(
                                        tgElem, HostExtType.class);

                                if (cur != null)
                                {
                                    HostExtType newHost = new HostExtType();

                                    if (cur.getName() != null)
                                    {
                                        String destHostName = cur.getName();
                                        if (args.length > 1)
                                        {
                                            destHostName += args[1];
                                        }
                                        newHost.setName(destHostName);
                                    }

                                    if (cur.getNodeProperties() != null)
                                    {
                                        newHost.setNodeProperties(
                                            cur.getNodeProperties());
                                    }

                                    newHostExts[j] = newHost;

                                    Element elem = ObjectSerializer.toElement(
                                        newHost,
                                        newHost.getTypeDesc().getXmlType());

                                    newMsgElems[j] = new MessageElement(elem);
                                }
                            }
                            else
                            {
                                newMsgElems[j] = new MessageElement(tgElem);
                            }
                        }

                        ce.set_any(newMsgElems);
                    }
                    catch (Exception e)
                    {
                        logger.error("Failed to retrieve the Any msgElements: "
                                     + e.toString());
                    }
                }

                String queueName = queue.getName();

                ce.setName(queueName);
                ce.setUniqueID(queueName);

                if (info != null)
                {
                    if (additionalInfo != null)
                    {
                        try
                        {
                            // override any values in the original
                            // info object that are non-null in the
                            // additionalInfo object
                            if (additionalInfo.getHostName() != null)
                                info.setHostName(additionalInfo.getHostName());
                            if (additionalInfo.getLRMSType() != null)
                                info.setLRMSType(additionalInfo.getLRMSType());
                            if (additionalInfo.getLRMSVersion() != null)
                                info.setLRMSVersion(additionalInfo.getLRMSVersion());
                            if (additionalInfo.getGRAMVersion() != null)
                                info.setGRAMVersion(additionalInfo.getGRAMVersion());
                            if (additionalInfo.getTotalCPUs() != null)
                                info.setTotalCPUs(additionalInfo.getTotalCPUs());
                            if (additionalInfo.getGatekeeperPort() != null)
                                info.setGatekeeperPort(additionalInfo.getGatekeeperPort());
                        }
                        catch(Exception e)
                        {
                            logger.debug(
                                "Failed to set InfoType Parameters " +
                                "from additionalInfo: " + e);
                        }
                    }

                    if (info.getTotalCPUs() == null)
                    {
                        try {
                            info.setTotalCPUs(
                                "" + queue.getTotalnodes().getIntValue());
                        }
                        catch(Exception e)
                        {
                            logger.debug("Failed to set InfoType " +
                                         "TotalCPUs: " + e);
                        }
                    }

                    if (info.getHostName() == null)
                    {
                        try {
                            info.setHostName(
                                java.net.InetAddress.getLocalHost().getCanonicalHostName());
                        }
                        catch(Exception e)
                        {
                            logger.debug("Failed to set InfoType " +
                                         "HostName: " + e);
                        }
                    }

                    ce.setInfo(info);
                }
                else if ((info == null) && (additionalInfo != null))
                {
                    info = additionalInfo;
                }

                StateType state = new StateType();
                state.setStatus(queue.getStatus());

/*
TODO: not sure how to map these fields from the provider schema:
<maxCount>143</maxCount>
<maxReqNodes>unlimited</maxReqNodes>
<maxTotalMemory>unlimited</maxTotalMemory>
<maxSingleMemory>unlimited</maxSingleMemory>
<whenActive>0</whenActive>
<dispatchType>batch</dispatchType>

Also, no mappings for these state fields?
private int worstResponseTime;  // attribute
private int estimatedResponseTime;  // attribute
*/

                try {
                    state.setFreeCPUs(queue.getFreenodes().getIntValue());
                } catch(Exception e) {
                    logger.debug("Unable to parse freeNodes value from queue information" +e);
                }

                int totalJobs = 0;
                try {
                    totalJobs = queue.getTotalJobs().getIntValue();                    
                    state.setTotalJobs(totalJobs);
                } catch(Exception e) {
                    logger.debug("Unable to parse totalJobs value from queue information" +e);
                }

                int runningJobs = 0;
                try {
                    runningJobs = queue.getRunningJobs().getIntValue();                    
                    state.setRunningJobs(runningJobs);
                } catch(Exception e) {
                    logger.debug("Unable to parse runningJobs value from queue information" +e);
                }

                int waitingJobs = (totalJobs - runningJobs);
                state.setWaitingJobs(waitingJobs);

                ce.setState(state);

                PolicyType policy = new PolicyType();

                try {
                    policy.setMaxWallClockTime(queue.getMaxtime().getIntValue());
                } catch(Exception e) {
                    logger.debug("Unable to parse maxTime value from queue information" +e);
                    policy.setMaxWallClockTime(-1);
                }

                try {
                    policy.setMaxCPUTime(queue.getMaxCPUtime().getIntValue());
                } catch(Exception e) {
                    logger.debug("Unable to parse maxCPUtime value from queue information" +e);
                    policy.setMaxCPUTime(-1);
                }

                try {
                    policy.setMaxRunningJobs(queue.getMaxRunningJobs().getIntValue());
                } catch(Exception e) {
                    logger.debug("Unable to parse maxRunningJobs value from queue information" +e);                    
                    policy.setMaxRunningJobs(-1);
                }

                try {
                    policy.setMaxTotalJobs(queue.getMaxJobsInQueue().getIntValue());
                } catch(Exception e) {
                    logger.debug("Unable to parse maxJobsInQueue value from queue information" +e);                    
                    policy.setMaxTotalJobs(-1);
                }

/*
No mappings for these policy attributes:
private int priority;  // attribute
*/
                ce.setPolicy(policy);            
                ce.setJob(jobs);

                ces[i] = ce;
            }
        }
        catch(Exception e) 
        {
            logger.warn("Unhandled exception during GLUE ComputeElement transformation",e);
        }
        return ces;
    }     
}
    
