/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
* 
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider.transforms;

import java.util.*;

import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import javax.xml.namespace.QName;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.container.ServiceHost;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertyMetaData;

import org.globus.wsrf.encoding.DeserializationException;

import org.w3c.dom.Node;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.mds.glue.MainMemoryType;
import org.globus.mds.glue.ProcessorType;
import org.globus.mds.glue.OperatingSystemType;
import org.globus.mds.glue.ArchitectureType;
import org.globus.mds.glue.StorageDeviceType;

import org.globus.mds.glue.config.ClusterConfiguration;
import org.globus.mds.glue.config.ClusterDefinitionType;
import org.globus.mds.glue.config.SubClusterDefinitionType;
import org.globus.mds.glue.config.HostNameArrayDefinitionType;
import org.globus.mds.glue.config.HostNameDefinitionType;


public class GLUEXMLOrganizer
{
    private static final String NAMESPACE_XMLNS =
        "http://www.w3.org/2000/xmlns/";
    private static final String NAMESPACE_GLUE =
        "http://mds.globus.org/glue/ce/1.1";
    private static final String NAMESPACE_CLUSTERCFG =
        "http://mds.globus.org/2005/09/cluster-config";

    private static Log logger =
            LogFactory.getLog(GLUEXMLOrganizer.class.getName());

    // custom processing states
    private static final int NONE = 2;
    private static final int ADD_TO_HOST_FALSE = 3;
    private static final int CONTINUE = 4;

    // values representing the element types we can handle
    private static final int UNKNOWN = 6;
    private static final int MAIN_MEMORY = 7;
    private static final int PROCESSOR = 8;
    private static final int OPERATING_SYSTEM = 9;
    private static final int ARCHITECTURE = 10;
    private static final int STORAGEDEVICE = 11;

    public static ClusterConfiguration parseConfig(String configFile)
    {
        ClusterConfiguration config = null;

        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setIgnoringComments(true);
        factory.setCoalescing(true);
        factory.setNamespaceAware(true);
        factory.setValidating(false);

        try
        {
            DocumentBuilder db = factory.newDocumentBuilder();
            Document doc = db.parse(new File(configFile));
            Element elem = doc.getDocumentElement();

            config = (ClusterConfiguration)ObjectDeserializer.toObject(
                elem, ClusterConfiguration.class);

            if (config != null)
            {
                if (config.getCluster() == null)
                {
                    logger.error(
                        "Cluster configuration information is invalid.");
                }
                else if (config.getCluster().getSubCluster() == null)
                {
                    logger.error(
                        "SubCluster configuration information is invalid.");
                }
            }
        }
        catch(Exception e)
        {
            logger.error("Failed to load configuration: " + e);
        }
        return config;
    }

    private static String replaceNS(String attr)
    {
        String newAttr = attr;
        if (attr != null)
        {
            int colonPos = attr.indexOf(':');
            if (colonPos != -1)
            {
                newAttr = "glue:" + attr.substring(colonPos+1);
            }
        }
        return newAttr;
    }

    public static String removeNS(String attr)
    {
        String newAttr = attr;
        if (attr != null)
        {
            int colonPos = attr.indexOf(':');
            if (colonPos != -1)
            {
                newAttr = attr.substring(colonPos+1);
            }
        }
        return newAttr;
    }

    // we're keeping track of two maps here:
    // the obj map maps a host name to the subCluster type object
    // the elem map maps a host name to the subCluster element object
    private static void mapHostsToSubCluster(
        Map hostObjMap, Map hostElemMap, Map hostUIDMap,
        Element subClusterElem, SubClusterDefinitionType subClusterObj,
        HostNameArrayDefinitionType hostnames)
    {
        if ((hostObjMap != null) && (hostElemMap != null) &&
            (subClusterElem != null) && (subClusterObj != null) &&
            (hostnames != null))
        {
            int numHostNames = java.lang.reflect.Array.getLength(
                hostnames.getHostName());
            for(int i = 0; i < numHostNames; i++)
            {
                HostNameDefinitionType curHost = hostnames.getHostName(i);
                if ((curHost != null) && (curHost.getName() != null))
                {
                    String currentHostStr = curHost.getName();

                    hostObjMap.put(currentHostStr, subClusterObj);
                    hostElemMap.put(currentHostStr, subClusterElem);
                    hostUIDMap.put(currentHostStr, curHost);
                }
            }
        }
    }

    private static int checkMainMemoryForSubClusterLevel(
        MainMemoryType mem, Map addedToSubClusterDupMap,
        SubClusterDefinitionType currentSubClusterObj)
    {
        int ret = NONE;

        if (mem == null)
        {
            return ret;
        }

        Map tmpMap = (Map)addedToSubClusterDupMap.get(currentSubClusterObj);
        if (tmpMap == null)
        {
            tmpMap = new HashMap();
        }

        if (tmpMap.get("MainMemory") != mem)
        {
/*             if ((mem.getRAMSize() > 0) || */
/*                 (mem.getRAMAvailable() > 0) || */
/*                 (mem.getVirtualSize() > 0) || */
/*                 (mem.getVirtualAvailable() > 0)) */
/*             { */
                ret = ADD_TO_HOST_FALSE;

                // NOTE: we're hashing to a hashmap since
                // the first lookup is the subcluster, and
                // the second is the specific element name
                // (e.g. MainMemory)
                tmpMap.put("MainMemory", mem);
                addedToSubClusterDupMap.put(
                    currentSubClusterObj, tmpMap);
/*             } */
        }
        else
        {
            ret = CONTINUE;
        }
        return ret;
    }

    private static int checkProcessorForSubClusterLevel(
        ProcessorType proc, Map addedToSubClusterDupMap,
        SubClusterDefinitionType currentSubClusterObj)
    {
        int ret = NONE;

        if (proc == null)
        {
            return ret;
        }

        Map tmpMap = (Map)addedToSubClusterDupMap.get(currentSubClusterObj);
        if (tmpMap == null)
        {
            tmpMap = new HashMap();
        }

        if (tmpMap.get("Processor") != proc)
        {
/*             if ((proc.getVendor() != null) || (proc.getModel() != null) || */
/*                 (proc.getVersion() != null) || */
/*                 (proc.getInstructionSet() != null) || */
/*                 (proc.getClockSpeed() > 0) || (proc.getCacheL1() > 0) || */
/*                 (proc.getCacheL1I() > 0) || (proc.getCacheL1D() > 0) || */
/*                 (proc.getCacheL2() > 0)) */
/*             { */
                ret = ADD_TO_HOST_FALSE;

                tmpMap.put("Processor", proc);
                addedToSubClusterDupMap.put(
                    currentSubClusterObj, tmpMap);
/*             } */
        }
        else
        {
            ret = CONTINUE;
        }
        return ret;
    }

    private static int checkOperatingSystemForSubClusterLevel(
        OperatingSystemType os, Map addedToSubClusterDupMap,
        SubClusterDefinitionType currentSubClusterObj)
    {
        int ret = NONE;

        if (os == null)
        {
            return ret;
        }

        Map tmpMap = (Map)addedToSubClusterDupMap.get(currentSubClusterObj);
        if (tmpMap == null)
        {
            tmpMap = new HashMap();
        }

        if (tmpMap.get("OperatingSystem") != os)
        {
/*             if ((os.getName() != null) || */
/*                 (os.getRelease() != null) || */
/*                 (os.getVersion() != null)) */
/*             { */
                ret = ADD_TO_HOST_FALSE;

                tmpMap.put("OperatingSystem", os);
                addedToSubClusterDupMap.put(
                    currentSubClusterObj, tmpMap);
/*             } */
        }
        else
        {
            ret = CONTINUE;
        }
        return ret;
    }

    private static int checkArchitectureForSubClusterLevel(
        ArchitectureType arch, Map addedToSubClusterDupMap,
        SubClusterDefinitionType currentSubClusterObj)
    {
        int ret = NONE;

        if (arch == null)
        {
            return ret;
        }

        Map tmpMap = (Map)addedToSubClusterDupMap.get(currentSubClusterObj);
        if (tmpMap == null)
        {
            tmpMap = new HashMap();
        }

        if (tmpMap.get("Architecture") != arch)
        {
/*             if ((arch.getPlatformType() != null) || */
/*                 (arch.getSMPSize() > 0)) */
/*             { */
                ret = ADD_TO_HOST_FALSE;

                tmpMap.put("Architecture", arch);
                addedToSubClusterDupMap.put(
                    currentSubClusterObj, tmpMap);
/*             } */
        }
        else
        {
            ret = CONTINUE;
        }
        return ret;
    }

    private static int checkStorageDeviceForSubClusterLevel(
        StorageDeviceType[] storageArray, Map addedToSubClusterDupMap,
        SubClusterDefinitionType currentSubClusterObj)
    {
        int ret = NONE;

        // NOTE: for now, we only handle a single StorageDevice object
        StorageDeviceType storage = ((storageArray == null) ?
                                     null : storageArray[0]);

        if (storage == null)
        {
            return ret;
        }

        Map tmpMap = (Map)addedToSubClusterDupMap.get(currentSubClusterObj);
        if (tmpMap == null)
        {
            tmpMap = new HashMap();
        }

        if (tmpMap.get("StorageDevice") != storage)
        {
/*             if ((storage.getName() != null) || */
/*                 (storage.getType() != null) || */
/*                 (storage.getTransferRate() > 0) || */
/*                 (storage.getSize() > 0) || */
/*                 (storage.getAvailableSpace() > 0)) */
/*             { */
                ret = ADD_TO_HOST_FALSE;

                tmpMap.put("StorageDevice", storage);
                addedToSubClusterDupMap.put(
                    currentSubClusterObj, tmpMap);
/*             } */
        }
        else
        {
            ret = CONTINUE;
        }
        return ret;
    }

    private static String getOrSetConvertedMemAttributeValue(
        String attrName, MainMemoryType mem, String hostValue)
    {
        String ret = null;
        boolean memIsNull = (mem == null) ? true : false;
        int len = hostValue.length();

        if (attrName.equals("RAMSize"))
        {
            if (memIsNull ||
                (!memIsNull && (mem.getRAMSize() == 0) && (len > 0)))
            {
                mem.setRAMSize(new Long(hostValue).longValue());
            }
            ret = String.valueOf(mem.getRAMSize());
        }
        else if (attrName.equals("RAMAvailable"))
        {
            if (memIsNull ||
                (!memIsNull && (mem.getRAMAvailable() == 0) && (len > 0)))
            {
                mem.setRAMAvailable(new Long(hostValue).longValue());
            }
            ret = String.valueOf(mem.getRAMAvailable());
        }
        else if (attrName.equals("VirtualSize"))
        {
            if (memIsNull ||
                (!memIsNull && (mem.getVirtualSize() == 0) && (len > 0)))
            {
                mem.setVirtualSize(new Long(hostValue).longValue());
            }
            ret = String.valueOf(mem.getVirtualSize());
        }
        else if (attrName.equals("VirtualAvailable"))
        {
            if (memIsNull ||
                (!memIsNull && (mem.getVirtualAvailable() == 0) && (len > 0)))
            {
                mem.setVirtualAvailable(new Long(hostValue).longValue());
            }
            ret = String.valueOf(mem.getVirtualAvailable());
        }
        return ret;
    }

    private static String getOrSetConvertedProcAttributeValue(
        String attrName, ProcessorType proc, String hostValue)
    {
        String ret = null;
        boolean procIsNull = (proc == null) ? true : false;
        int len = hostValue.length();

        if (attrName.equals("CacheL1"))
        {
            if (procIsNull ||
                (!procIsNull && (proc.getCacheL1() == 0) && (len > 0)))
            {
                proc.setCacheL1(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(proc.getCacheL1());
        }
        else if (attrName.equals("CacheL1I"))
        {
            if (procIsNull ||
                (!procIsNull && (proc.getCacheL1I() == 0) && (len > 0)))
            {
                proc.setCacheL1I(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(proc.getCacheL1I());
        }
        else if (attrName.equals("CacheL1D"))
        {
            if (procIsNull ||
                (!procIsNull && (proc.getCacheL1D() == 0) && (len > 0)))
            {
                proc.setCacheL1D(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(proc.getCacheL1D());
        }
        else if (attrName.equals("CacheL2"))
        {
            if (procIsNull ||
                (!procIsNull && (proc.getCacheL2() == 0) && (len > 0)))
            {
                proc.setCacheL2(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(proc.getCacheL2());
        }
        else if (attrName.equals("ClockSpeed"))
        {
            if (procIsNull ||
                (!procIsNull && (proc.getClockSpeed() == 0) && (len > 0)))
            {
                proc.setClockSpeed(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(proc.getClockSpeed());
        }
        else if (attrName.equals("InstructionSet"))
        {
            if (procIsNull || (proc.getInstructionSet() == null) ||
                (!procIsNull && (proc.getInstructionSet().length() == 0) &&
                 (len > 0)))
            {
                proc.setInstructionSet(hostValue);
            }
            ret = proc.getInstructionSet();
        }
        else if (attrName.equals("Model"))
        {
            if (procIsNull || (proc.getModel() == null) ||
                (!procIsNull && (proc.getModel().length() == 0) &&
                 (len > 0)))
            {
                proc.setModel(hostValue);
            }
            ret = proc.getModel();
        }
        else if (attrName.equals("Version"))
        {
            if (procIsNull || (proc.getVersion() == null) ||
                (!procIsNull && (proc.getVersion().length() == 0) &&
                 (len > 0)))
            {
                proc.setVersion(hostValue);
            }
            ret = proc.getVersion();
        }
        else if (attrName.equals("Vendor"))
        {
            if (procIsNull || (proc.getVendor() == null) ||
                (!procIsNull && (proc.getVendor().length() == 0) &&
                 (len > 0)))
            {
                proc.setVendor(hostValue);
            }
            ret = proc.getVendor();
        }
        return ret;
    }

    private static String getOrSetConvertedOsAttributeValue(
        String attrName, OperatingSystemType os, String hostValue)
    {
        String ret = null;
        boolean osIsNull = (os == null) ? true : false;
        int len = hostValue.length();

        if (attrName.equals("Name"))
        {
            if (osIsNull || (os.getName() == null) ||
                (!osIsNull && (os.getName().length() == 0) && (len > 0)))
            {
                os.setName(hostValue);
            }
            ret = os.getName();
        }
        else if (attrName.equals("Release"))
        {
            if (osIsNull || (os.getRelease() == null) ||
                (!osIsNull && (os.getRelease().length() == 0) && (len > 0)))
            {
                os.setRelease(hostValue);
            }
            ret = os.getRelease();
        }
        else if (attrName.equals("Version"))
        {
            if (osIsNull || (os.getVersion() == null) ||
                (!osIsNull && (os.getVersion().length() == 0) && (len > 0)))
            {
                os.setVersion(hostValue);
            }
            ret = os.getVersion();
        }
        return ret;
    }

    private static String getOrSetConvertedArchAttributeValue(
        String attrName, ArchitectureType arch, String hostValue)
    {
        String ret = null;
        boolean archIsNull = (arch == null) ? true : false;
        int len = hostValue.length();

        if (attrName.equals("PlatformType"))
        {
            if (archIsNull || (arch.getPlatformType() == null) ||
                (!archIsNull && (arch.getPlatformType().length() == 0) &&
                 (len > 0)))
            {
                arch.setPlatformType(hostValue);
            }
            ret = arch.getPlatformType();
        }
        else if (attrName.equals("SMPSize"))
        {
            if (archIsNull ||
                (!archIsNull && (arch.getSMPSize() == 0) && (len > 0)))
            {
                arch.setSMPSize(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(arch.getSMPSize());
        }
        return ret;
    }

    private static String getOrSetConvertedStorageDeviceAttributeValue(
        String attrName, StorageDeviceType[] storageArray, String hostValue)
    {
        String ret = null;
        // NOTE: for now, we only handle a single StorageDevice object
        StorageDeviceType storage = ((storageArray == null) ?
                                     null : storageArray[0]);
        boolean storageIsNull = (storage == null) ? true : false;
        int len = hostValue.length();

        if (attrName.equals("Name"))
        {
            if (storageIsNull || (storage.getName() == null) ||
                (!storageIsNull && (storage.getName().length() == 0) &&
                 (len > 0)))
            {
                storage.setName(hostValue);
            }
            ret = storage.getName();
        }
        else if (attrName.equals("Type"))
        {
            if (storageIsNull || (storage.getType() == null) ||
                (!storageIsNull && (storage.getType().length() == 0) &&
                 (len > 0)))
            {
                storage.setType(hostValue);
            }
            ret = storage.getType();
        }
        else if (attrName.equals("TransferRate"))
        {
            if (storageIsNull ||
                (!storageIsNull && (storage.getTransferRate() == 0) &&
                 (len > 0)))
            {
                storage.setTransferRate(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(storage.getTransferRate());
        }
        else if (attrName.equals("Size"))
        {
            if (storageIsNull ||
                (!storageIsNull && (storage.getSize() == 0) &&
                 (len > 0)))
            {
                storage.setSize(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(storage.getSize());
        }
        else if (attrName.equals("AvailableSpace"))
        {
            if (storageIsNull ||
                (!storageIsNull && (storage.getAvailableSpace() == 0) &&
                 (len > 0)))
            {
                storage.setAvailableSpace(new Integer(hostValue).intValue());
            }
            ret = String.valueOf(storage.getAvailableSpace());
        }
        return ret;
    }


    public static Document reorganizeGLUEDocument(
        ClusterConfiguration config, Document sourceDoc, String uidToAppend)
        throws Exception, IOException, ParserConfigurationException,
        org.xml.sax.SAXException
    {
        ClusterDefinitionType cluster = config.getCluster();
        SubClusterDefinitionType[] subCluster = cluster.getSubCluster();

        // start constructing the destination XML document
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setIgnoringComments(true);
        factory.setCoalescing(true);
        factory.setNamespaceAware(true);
        factory.setValidating(false);

        DocumentBuilder db = factory.newDocumentBuilder();
        Document destDoc = db.newDocument();


        Element root = destDoc.createElementNS(NAMESPACE_GLUE, "glue:ClusterInformation");
        destDoc.appendChild(root);

        Element destCluster = destDoc.createElementNS(
            NAMESPACE_GLUE, "glue:Cluster");
        destCluster.setAttributeNS(
            NAMESPACE_GLUE, "glue:Name", cluster.getName());
        destCluster.setAttributeNS(
            NAMESPACE_GLUE, "glue:UniqueID", cluster.getUniqueID());
        root.appendChild(destCluster);

        int i = 0;
        int numSubClusters = java.lang.reflect.Array.getLength(subCluster);

        Map hostToSubClusterObjMap = new HashMap();
        Map hostToSubClusterElemMap = new HashMap();
        Map hostToUniqueIDMap = new HashMap();
        Map subClusterToHostElemMap = new HashMap();

        // initialize the hash map for tracking duplicate host
        // attribute elements that we've already bubbled up to the
        // SubCluster level.  a lookup on this actually hashes to
        // another hashMap
        Map addedToSubClusterDupMap = new HashMap();

        HostNameArrayDefinitionType hostNameArray = null;
        for (i = 0; i < numSubClusters; i++)
        {
            SubClusterDefinitionType cur = cluster.getSubCluster(i);
            Element destSubCluster = destDoc.createElementNS(
                NAMESPACE_GLUE, "glue:SubCluster");

            destSubCluster.setAttributeNS(
                NAMESPACE_GLUE, "glue:Name", cur.getName());
            destSubCluster.setAttributeNS(
                NAMESPACE_GLUE, "glue:UniqueID", cur.getUniqueID());
            destCluster.appendChild(destSubCluster);

            hostNameArray = cur.getHostNames();
            if (hostNameArray == null)
            {
                throw new Exception(
                    "No hosts specified.  Nothing to transform.");
            }
            mapHostsToSubCluster(hostToSubClusterObjMap,
                                 hostToSubClusterElemMap,
                                 hostToUniqueIDMap,
                                 destSubCluster, cur,
                                 hostNameArray);
        }

        Element currentSubClusterElem = null;
        SubClusterDefinitionType currentSubClusterObj = null;

        // initialize the hash map for tracking duplicate host element
        // entries within the input GLUE data
        Map duplicateHostMap = new HashMap();

        Element firstHost = null;

        NodeList hosts = sourceDoc.getElementsByTagNameNS(NAMESPACE_GLUE, "Host");
        int numHosts = hosts.getLength();

        // add total host count to top-level cluster element
        Element totalNodes = destDoc.createElementNS(
            NAMESPACE_CLUSTERCFG, "cfg:TeraGridExtensions");
        totalNodes.setAttributeNS(
            NAMESPACE_CLUSTERCFG, "cfg:TotalNodes",
            new Integer(numHosts).toString());
        destCluster.appendChild(totalNodes);

        for(i = 0; i < numHosts; i++)
        {
            Element host = (Element)hosts.item(i);
            String currentHostName = "";

            // add all host attributes from the source to the
            // destination document
            Element destHost = destDoc.createElementNS(NAMESPACE_GLUE, "glue:Host");
            NamedNodeMap srcAttrs = ((Node)host).getAttributes();
            for(int l = 0; l < srcAttrs.getLength(); l++)
            {
                Attr attr = (Attr)srcAttrs.item(l);
                String currentAttrName = replaceNS(attr.getName());
                HostNameDefinitionType currentHost = null;

                if (removeNS(attr.getName()).equals("Name"))
                {
                    currentHostName = attr.getValue();

                    if (uidToAppend != null)
                    {
                        currentHostName += uidToAppend;
                    }
                }

                // check if we're to override the UniqueID with the
                // value in the config file (assuming host name is
                // already set)
                if (removeNS(attr.getName()).equals("UniqueID") &&
                    (currentHostName != null))
                {
                    currentHost = (HostNameDefinitionType)
                        hostToUniqueIDMap.get(currentHostName);
                    if ((currentHost != null) &&
                        (currentHost.getUniqueID() != null))
                    {
                        destHost.setAttributeNS(
                            NAMESPACE_GLUE, currentAttrName,
                            currentHost.getUniqueID());
                    }
                    else
                    {
                        destHost.setAttributeNS(
                            NAMESPACE_GLUE, currentAttrName, attr.getValue());
                    }
                }
                else if ((removeNS(attr.getName()).equals("Name")) &&
                         (currentHostName != null))
                {
                    destHost.setAttributeNS(
                        NAMESPACE_GLUE, currentAttrName, currentHostName);
                }
                else
                {
                    destHost.setAttributeNS(
                        NAMESPACE_GLUE, currentAttrName, attr.getValue());
                }
            }

            // be sure to skip duplicate hosts, as it appears that the
            // DefaultIndexService can contain duplicates of the GLUE
            // data
            if (duplicateHostMap.containsKey(currentHostName))
            {
                continue;
            }
            duplicateHostMap.put(currentHostName, "1");

            currentSubClusterElem = (Element)
                hostToSubClusterElemMap.get(currentHostName);
            currentSubClusterObj = (SubClusterDefinitionType)
                hostToSubClusterObjMap.get(currentHostName);

            if ((currentSubClusterElem == null) ||
                (currentSubClusterObj == null))
            {
                // NOTE: we should just probably leave this host in a
                // default SubCluster area if it wasn't associated
                // with a SubCluster explicitly
                logger.error("Fatal error!  Unmapped hostname \"" +
                             currentHostName + "\" in GLUE input data.");
                logger.error("A host name was encountered that was not "
                             + "specified in the configuration.");
                logger.error("Make sure all hosts are part of a SubCluster "
                             + "definition.  Aborting");

                throw new Exception(
                    "Invalid configuration.  Check the error log for detail.");
            }

            // add this host to the subClusterToHostElemMap List based
            // on the current subcluster for later addition.  this
            // guarantees all bubbled up attributes will appear before
            // any host elements (important for deserialization)
            Vector curList = (Vector)
                subClusterToHostElemMap.get(currentSubClusterElem);
            if (curList == null)
            {
                curList = new Vector();
                subClusterToHostElemMap.put(currentSubClusterElem, curList);
            }
            curList.add(destHost);

            // and subsequently add all child elements of the host
            Node child = host.getFirstChild();
            while(child != null)
            {
                while((child != null) &&
                      (child.getNodeType() != Node.ELEMENT_NODE))
                {
                    child = child.getNextSibling();
                }

                if (child == null)
                {
                    break;
                }

                String childName = replaceNS(child.getNodeName());
                String cleanChildName = removeNS(childName);
                Element childElem = destDoc.createElementNS(
                    NAMESPACE_GLUE, childName);
                boolean addToHost = true;
                int processingState = NONE;
                int elementType = UNKNOWN;
                MainMemoryType mem = currentSubClusterObj.getMainMemory();
                ProcessorType proc = currentSubClusterObj.getProcessor();
                OperatingSystemType os = currentSubClusterObj.getOperatingSystem();
                ArchitectureType arch = currentSubClusterObj.getArchitecture();
                StorageDeviceType[] storage = currentSubClusterObj.getStorageDevice();

                // check here if this is a bubbled up attribute and
                // add it to the currentSubCluster level rather than
                // the host level.  we also have to check for
                // duplicates here since each host will have the same
                // attr that we're bubbling up
                if (cleanChildName.equals("MainMemory"))
                {
                    processingState = checkMainMemoryForSubClusterLevel(
                        mem, addedToSubClusterDupMap, currentSubClusterObj);

                    elementType = MAIN_MEMORY;
                }
                else if (cleanChildName.equals("Processor"))
                {
                    processingState = checkProcessorForSubClusterLevel(
                        proc, addedToSubClusterDupMap, currentSubClusterObj);

                    elementType = PROCESSOR;
                }
                else if (cleanChildName.equals("OperatingSystem"))
                {
                    processingState = checkOperatingSystemForSubClusterLevel(
                        os, addedToSubClusterDupMap, currentSubClusterObj);

                    elementType = OPERATING_SYSTEM;
                }
                else if (cleanChildName.equals("Architecture"))
                {
                    processingState = checkArchitectureForSubClusterLevel(
                        arch, addedToSubClusterDupMap, currentSubClusterObj);

                    elementType = ARCHITECTURE;
                }
                else if (cleanChildName.equals("StorageDevice"))
                {
                    processingState = checkStorageDeviceForSubClusterLevel(
                        storage, addedToSubClusterDupMap, currentSubClusterObj);

                    elementType = STORAGEDEVICE;
                }

                if (processingState == ADD_TO_HOST_FALSE)
                {
                    addToHost = false;
                }
                else if (processingState == CONTINUE)
                {
                    // if we've already bubbled this element up to
                    // the subCluster level, we can ignore it
                    // completely this time around and move on to
                    // processing the next element
                    child = child.getNextSibling();
                    continue;
                }

                int numInvalidAttrs = 0;
                NamedNodeMap childAttrs = child.getAttributes();
                for(int m = 0; m < childAttrs.getLength(); m++)
                {
                    Attr cAttr = (Attr)childAttrs.item(m);
                    String cAttrCleanName = removeNS(cAttr.getName());

                    if (addToHost == false)
                    {
                        String val = "";

                        if (elementType == MAIN_MEMORY)
                        {
                            val = getOrSetConvertedMemAttributeValue(
                                cAttrCleanName, mem, cAttr.getValue());
                        }
                        else if (elementType == PROCESSOR)
                        {
                            val = getOrSetConvertedProcAttributeValue(
                                cAttrCleanName, proc, cAttr.getValue());
                        }
                        else if (elementType == OPERATING_SYSTEM)
                        {
                            val = getOrSetConvertedOsAttributeValue(
                                cAttrCleanName, os, cAttr.getValue());
                        }
                        else if (elementType == ARCHITECTURE)
                        {
                            val = getOrSetConvertedArchAttributeValue(
                                cAttrCleanName, arch, cAttr.getValue());
                        }
                        else if (elementType == STORAGEDEVICE)
                        {
                            val = getOrSetConvertedStorageDeviceAttributeValue(
                                cAttrCleanName, storage, cAttr.getValue());
                        }
                        else
                        {
                            val = cAttr.getValue();
                        }

                        // don't emit fields that weren't specified
                        // (or are invalid)
                        if ((val == null) || val.equals("0"))
                        {
                            continue;
                        }
                        childElem.setAttributeNS(
                            NAMESPACE_GLUE, replaceNS(cAttr.getName()), val);
                    }
                    else
                    {
                        if ((cAttr.getValue() == null) ||
                            cAttr.getValue().equals(""))
                        {
                            numInvalidAttrs++;
                            continue;
                        }

                        // simply use the host value unchanged
                        childElem.setAttributeNS(
                            NAMESPACE_GLUE, replaceNS(cAttr.getName()),
                            cAttr.getValue());
                    }
                }

                // add this element to either the host or the current
                // subcluster element so long as it's not invalid.  by
                // invalid, we mean that it must have at least one
                // valid attribute within the element.  if all
                // attribute values are invalid in the element, the
                // element type cannot be properly (de)serialized.  to
                // avoid that future error, we're simply removing the
                // element here.
                if (childAttrs.getLength() != numInvalidAttrs)
                {
                    if (addToHost == true)
                    {
                        destHost.appendChild(childElem);
                    }
                    else
                    {
                        currentSubClusterElem.appendChild(childElem);
                    }
                }

                child = child.getNextSibling();
            }
        }

        // iterate over the subclusters and emit the Host and
        // TotalNodes elements
        Iterator keyIter = subClusterToHostElemMap.keySet().iterator();
        Iterator valIter = subClusterToHostElemMap.values().iterator();

        while (keyIter.hasNext() && valIter.hasNext())
        {
            Element subClusterElem = (Element)keyIter.next();
            Vector curHostElemList = (Vector)valIter.next();

            // for each host element in the list, add them to this
            // subcluster now
            int listSize = curHostElemList.size();
            for(i = 0; i < listSize; i++)
            {
                Element curHostElem = (Element)curHostElemList.get(i);
                subClusterElem.appendChild(curHostElem);
            }

            Element curTotalNodes = destDoc.createElementNS(
                NAMESPACE_CLUSTERCFG, "cfg:TeraGridExtensions");
            curTotalNodes.setAttributeNS(
                NAMESPACE_CLUSTERCFG, "cfg:TotalNodes",
                new Integer(listSize).toString());

            subClusterElem.appendChild(curTotalNodes);
        }

        if (logger.isDebugEnabled())
        {
            logger.debug("** Final Transformed document (before deserialization) ***");

            String output = ObjectSerializer.toString(
                cluster, cluster.getTypeDesc().getXmlType());
            logger.debug(output);
        }          

        return destDoc;
    }

    public static void transformDocument(Document doc)
    {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();

            //transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.ENCODING,"ISO-8859-1");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            DOMSource source = new DOMSource(doc);
            StreamResult result = new StreamResult(System.out);

            transformer.transform(source, result);
        }
        catch(Exception e)
        {
            logger.error("transformation of document failed!");
        }
    }
}
