/*
* Portions of this file Copyright 1999-2005 University of Chicago
* Portions of this file Copyright 1999-2005 The University of Southern California.
*
* This file or a portion of this file is licensed under the
* terms of the Globus Toolkit Public License, found at
* http://www.globus.org/toolkit/download/license.html.
* If you redistribute this file, with or without
* modifications, you must include this notice in the file.
*/
package org.globus.mds.usefulrp.rpprovider.transforms;

import org.globus.wsrf.utils.XmlUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.InputStream;
import java.io.FileInputStream;

import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.dom.DOMResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;

import org.globus.mds.usefulrp.rpprovider.ResourcePropertyElementTransform;

/** 
*/

public class XSLTFileElementTransform 
    implements ResourcePropertyElementTransform
{
    private static Log logger =
            LogFactory.getLog(XSLTFileElementTransform.class.getName());

    /** Creates a new instance of XSLTFileElementTransform */
    public XSLTFileElementTransform() {
    }

    public Object transformElement(Element source, String[] args) 
        throws Exception
    {
        if ((args == null) || ((args != null) && (args.length == 0))) {
             throw new IllegalArgumentException("Arguments cannot be null!");
        }
        
        InputStream xsltInputStream = null; 
        Element resultElement = null;

        try 
        {
            // set up transformer
            DOMResult result = new DOMResult();
            DOMSource domSource = new DOMSource(source);
            xsltInputStream = new FileInputStream(args[0]);
            StreamSource xsltSource = new StreamSource(xsltInputStream);
            TransformerFactory transformerFactory= TransformerFactory.newInstance();
            Transformer transformer = transformerFactory.newTransformer(xsltSource);

            // transform!
            transformer.transform(domSource,result);

            // extract result Element
            Document resultNode = (Document)result.getNode();
            resultElement = resultNode.getDocumentElement();           
        }
        catch (Exception e)
        {
            throw new Exception("Unhandled exception during XSLT transform");
        }
        finally
        {
            if (xsltInputStream != null) {
                xsltInputStream.close();
            }
        }
        
        return resultElement;
    }
}
